//
// C++ Implementation: Remote table model
//
//
// Author: Witold Wysota <wysota@wysota.eu.org>, (C) 2006
//
//
//
#include "remotetablemodel.h"
#include <QTimer>
#include <QEvent>
#include <QCoreApplication>
#include <QThread>


class UpdateEvent : public QEvent {
public:
    enum UpType { RetrieveAll, RetrieveOne, UpdateOne, InsertOne, RemoveOne };
    UpdateEvent() : QEvent((Type)20987) {
        init();
        m_type = RetrieveAll;
    }
    UpdateEvent(UpType t, int id) : QEvent((Type)20987) {
        init();
        m_type = t;
        m_id = id;
    }
    UpdateEvent(int id, int f, const QVariant &v, int role=Qt::EditRole) : QEvent((Type)20987) {
        init();
        m_type = UpdateOne;
        m_id = id;
        m_field = f;
        m_value = v;
        m_role = role;
    }
    UpType uptype() const {
        return m_type;
    }
    int id() const {
        return m_id;
    }
    int field() const {
        return m_field;
    }
    const QVariant &value() const {
        return m_value;
    }
    int role() const {
        return m_role;
    }
private:
    void init() {
        m_id = -1;
        m_field = -1;
        m_role = -1;
    }
    UpType m_type;
    int m_id;
    int m_field;
    QVariant m_value;
    int m_role;
};

/**
 *  Inserts a new row into the storage with the \a rowid id
 *
 *  Base implementation does nothing.
 */
Updater insertRow( int )
{
}

/**
 *  Removes a row with id \a rowid from the remote storage
 *
 *  Base implementation does nothing.
 */
Updater removeRow( int )
{
}

/**
 *  Updates \a rowid and \a column item \a role with \a data
 *
 *  Base implementation does nothing.
 */
Updater setData( int , int , const QVariant &, int  )
{
}

/**
 *  Retrieves a single row with id \a rowid from the remote storage.
 *
 *  Method should emit the retrieved() signal after getting the row.
 *
 *  Base implementation does nothing.
 */
Updater retrieveRow( int )
{
}

/**
 *  Retrieves all rows from the remote storage.
 *
 *  Method should emit the retrieved() signal after getting each row.
 *
 *  Base implementation does nothing.
 */
void Updater::retrieveAll( )
{
}


void Updater::customEvent( QEvent * event ) {
    if(event->type()!=20987)
        return QObject::customEvent(event);
    UpdateEvent *update = (UpdateEvent*)event;
    switch(update->uptype()) {
    case UpdateEvent::RetrieveAll:
        retrieveAll();
        update->accept();
        return;
    case UpdateEvent::RetrieveOne:
        retrieveRow(update->id());
        update->accept();
        return;
    case UpdateEvent::UpdateOne:
        setData(update->id(), update->field(), update->value(), update->role());
        update->accept();
        return;
    case UpdateEvent::InsertOne:
        insertRow(update->id());
        update->accept();
        return;
    case UpdateEvent::RemoveOne:
        removeRow(update->id());
        update->accept();
        return;
    }
    update->ignore();

}

class UpdaterThread : public QThread {
Q_OBJECT
public:
  UpdaterThread(QObject *parent=0);
  ~UpdaterThread();
  void setUpdaterFactory(UpdaterFactory *up);
public slots:
  void retrieveAll();
  void setData(int rowid, int column, const QVariant &data, int role=Qt::EditRole);
  void insertRow(int rowid);
  void removeRow(int rowid);
signals:
  void retrieved(const QVariant &data);
  void ready();
protected:
  void run(); // initialise objects and call exec()
private:
  Updater *m_updater;
  UpdaterFactory *m_updaterfactory;
};

#include "remotetablemodel.moc"

UpdaterThread::UpdaterThread( QObject * parent ) : QThread(parent) {
    m_updater = 0;
    m_updaterfactory = 0;
}

UpdaterThread::~UpdaterThread() {
    delete m_updater;
}


void UpdaterThread::retrieveAll( ) {
    QCoreApplication::postEvent(m_updater, new UpdateEvent());
}


void UpdaterThread::setData( int rowid, int column, const QVariant & data, int role ) {
    QCoreApplication::postEvent(m_updater, new UpdateEvent(rowid, column, data, role));
}

void UpdaterThread::removeRow( int rowid ) {
    QCoreApplication::postEvent(m_updater, new UpdateEvent(UpdateEvent::RemoveOne, rowid));
}

void UpdaterThread::insertRow( int rowid ) {
    QCoreApplication::postEvent(m_updater, new UpdateEvent(UpdateEvent::InsertOne, rowid));
}

void UpdaterThread::setUpdaterFactory( UpdaterFactory * up ) {
    m_updaterfactory = up;
}

void UpdaterThread::run( ) {
    if(!m_updaterfactory)
        return;
    m_updater = m_updaterfactory->create();
    connect(m_updater, SIGNAL(retrieved( const QVariant& )), SIGNAL(retrieved(const QVariant &)));
    emit ready();
    exec();
}


class RemoteTableModelPrivate : public QObject {
public:
  RemoteTableModelPrivate(QObject *parent) : QObject(parent){}
  ~RemoteTableModelPrivate(){
    m_thread.quit();
    m_thread.wait();
  }
  RemoteTableModel* q() const { return (RemoteTableModel*)parent(); }
  UpdaterThread m_thread;
};

/************************** PUBLIC API ********************/

/**
 *
 */
RemoteTableModel::RemoteTableModel( QObject * parent ) : QAbstractTableModel(parent) {
    m_d = new RemoteTableModelPrivate(this);
    qRegisterMetaType<QVariant>("QVariant");
    connect(&d()->m_thread, SIGNAL(ready()), SLOT(beginFill()));
    connect(&d()->m_thread, SIGNAL(retrieved( const QVariant& )), SLOT(addRow(const QVariant &)));
}

void RemoteTableModel::beginFill(){ d()->m_thread.retrieveAll(); }

/**
 *  Starts the updater and retrieves model contents from remote storage.
 */
void RemoteTableModel::start( ) {
    d()->m_thread.start();
}

/**
 *
 */
void RemoteTableModel::setUpdaterFactory(UpdaterFactory *up) {
    d()->m_thread.setUpdaterFactory(up);
}

/**
 *  Call from setData() to update remote model.
 */
void RemoteTableModel::remoteSetData(int rowid, int column, const QVariant &data, int role) {
    d()->m_thread.setData(rowid, column, data, role);
}

/**
 *  Call from insertRows() to insert a row into remote model.
 */
void RemoteTableModel::remoteInsertRow(int rowid) {
    d()->m_thread.insertRow(rowid);
}

/**
 *  Call from removeRows() to remove a row from remote model.
 */
void RemoteTableModel::remoteRemoveRow(int rowid) {
    d()->m_thread.removeRow(rowid);
}
